#!/usr/bin/env python

import base64
from binascii import hexlify
import cgitb
from contextlib import contextmanager
import fnmatch
from functools import partial
from functools import wraps
import inspect
import hashlib
import hmac
import json
import logging
import os
import re
import requests
import shutil
import socket
import string
import sys
import tempfile
import threading
import time
import types
import zipfile
import urllib.parse
try:
    from scandir import walk
except ImportError:
    from os import walk

from werkzeug.exceptions import BadRequest, Forbidden

from .const import OAUTH_BASE_URL, EDU_EOL_DATES, TIMEOUT
from .compat import RegistryManager, consteq, request_jsonrequest, request_type
from .auth_util import check_message_signature, extract_signature_from_kwargs, SaasAuth
from . import metabase

from odoo import api, release, SUPERUSER_ID
from odoo.http import request, root
from odoo.osv import expression
from odoo.tools import exception_to_unicode, reverse_enumerate, config, ustr
from odoo.tools.parse_version import parse_version


_logger = logging.getLogger(__name__)
HERE = os.path.dirname(os.path.realpath(__file__))
_COUNTRIES = None

GEOIP_FALLBACK = {
    'country_code': 'BE',
    'city': 'Unlocalized',
}

try:
    from odoo.tools import cloc
    cloc.STANDARD_MODULES += [
        'saas_demo',
        'saas_trial',
        'openerp_enterprise',
        'theme_common',
        'openerp_test',
        'saas_worker',
    ]
except Exception:
    pass

def check_dbname(dbname):
    return dbname != 'meta' and re.match(r"^[a-z][a-z0-9\-]{2,}[a-z0-9]$", dbname)

def version_gte(version):
    return parse_version(release.series) >= parse_version(version)

def version_lte(version):
    return parse_version(release.series) <= parse_version(version)


if version_gte("saas~16.1"):
    from odoo.http import GeoIP
    def ip2coordinates(ip):
        # https://github.com/odoo/odoo/commit/c59750d8244068e191a78631f8e0fd3d09067e3d
        result = GeoIP(ip)
        return {
            'city': result.city.name,
            'country_code': result.country.iso_code or result.continent.code,
            'country_name': result.country.name or result.continent.name,
            'latitude': result.location.latitude,
            'longitude': result.location.longitude,
            'region': result.subdivisions[0].iso_code if result.subdivisions else None,
            'time_zone': result.location.time_zone,
        } if result else GEOIP_FALLBACK
else:
    try:
        try:
            from odoo.tools.geoipresolver import GeoIPResolver
        except ImportError:
            from odoo.addons.http_routing.geoipresolver import GeoIPResolver
        geofile = config.get('geoip_database', '/usr/share/GeoIP/GeoLite2-City.mmdb')
        _geo_ip = GeoIPResolver.open(geofile)
        assert _geo_ip

        def ip2coordinates(ip):
            result = _geo_ip.resolve(ip)
            if result:
                for i in result:
                    if isinstance(result[i], bytes):
                        result[i] = result[i].decode('latin1')
            return result if (result and result.get('country_code')) else GEOIP_FALLBACK

    except Exception:
        _logger.exception('GeoIP: general failure')

        def ip2coordinates(ip):     # noqa
            return GEOIP_FALLBACK

def oe_branch(version, short=False):
    """Returns the branch associated to a version number.

    Args:
        version (int): version number.
        short (bool): toggle short mode, defaults to False.

    Examples:
        >>> print(oe_branch(100015, short=True))
        'saas~15'

        >>> print(oe_branch(100015))
        '10.saas~15'
    """
    minor = version % 100
    major = int(version / 1000) if minor < 10 else int(version / 10000)
    if minor == 0:  # ^\d+\.0
        fmt = "{major}.{minor}"
    elif major <= 10:  # ^\d+\.saas-\d+
        if short:
            fmt = "saas~{minor}"
        else:
            fmt = "{major}.saas~{minor}"
    else: # ^saas-\d+\.\d+
        fmt = "saas~{major}.{minor}"
    return fmt.format(major=major, minor=minor)

def humanize_version(version):
    """ human readable version
        6.1
        10.saas~15+e    -> saas-15
        11.0+e          -> 11.0
        10.0-20180507   -> 10.0
        saas~11.2       -> saas-11.2
        saas-12.3       -> saas-12.3
    """
    version = version.lower()
    # some Windows builds have a double `e` in the series, e.g. 12.0e+e instead of 12.0+e
    version = version.replace("e+e", "+e")
    v_re = r'(\d+(?:\.\d+))|(saas(~|-)\d+(?:\.\d+){0,1})'
    match = re.search(v_re, version)
    if not match:
        return version
    groups = match.groups()
    return groups[1].replace('~', '-') if groups[1] else groups[0]


def get_edu_eol_date(version):
    if not version:
        return False
    version_info = humanize_version(version)
    return EDU_EOL_DATES.get(version_info, False)

class SupportLevel(object):
    CUSTOMER = 0
    EMPLOYEE = 50
    SALESMAN = 100
    CONSULTANT = 200
    FUNCTIONAL_SUPPORT = 300
    TECHNICAL_SUPPORT = 400


@contextmanager
def tempdir(dir=None):
    tmpdir = tempfile.mkdtemp(dir=dir)
    try:
        yield tmpdir
    finally:
        shutil.rmtree(tmpdir)


def zip_dir(path, include=True, ignore=(), append_to=None):
    path = os.path.normpath(path)
    match_prefix = len(path) + 1
    len_prefix = len(os.path.dirname(path)) if include else len(path)
    if len_prefix:
        len_prefix += 1

    ignore = ('*.py[co]', '*.swp', '**/.DS_Store') + tuple(ignore)

    def match(f):
        for pat in ignore:
            p = './' if pat.startswith('**/') and '/' not in f else ''
            if fnmatch.fnmatch(p + f, pat):
                return False
        return True

    zipio = append_to or tempfile.TemporaryFile()
    with zipfile.ZipFile(zipio, 'a', compression=zipfile.ZIP_DEFLATED, allowZip64=True) as zipf:
        for dirpath, dirnames, filenames in os.walk(path):
            match_dir = dirpath[match_prefix:]
            for i, d in reverse_enumerate(dirnames):
                if not match(os.path.join(match_dir, d)):
                    del dirnames[i]
            for fname in filenames:
                path = os.path.normpath(os.path.join(dirpath, fname))
                local_path = path[len_prefix:]
                if os.path.isfile(path) and match(local_path):
                    zipf.write(path, local_path)

    zipio.seek(0)
    return zipio

def get_size(dbname):
    # Dump size
    bakdir = os.path.expanduser('~/bakdb')
    backup_size = 0
    try:
        for backup_file in os.listdir(bakdir):
            # backups files have the file name pattern:
            # <dname>.<uuid>.<date>.sql.gz
            if backup_file.startswith(dbname + '.') and backup_file.endswith('.sql.gz'):
                backup_size = os.path.getsize(os.path.join(bakdir, backup_file))
                break
    except OSError:
        pass

    # Filestore size
    fs_size = sum(os.path.getsize(os.path.join(dirpath, filename))
                  for dirpath, _, filenames in walk(config.filestore(dbname))
                  for filename in filenames)

    return(backup_size, fs_size)


def get_country_details(country_code):
    global _COUNTRIES
    if _COUNTRIES is None:
        # json file taken from django-countries-plus
        raw = json.load(open(os.path.join(HERE, 'countries.json')))
        _COUNTRIES = dict((c['pk'], c['fields']) for c in raw)
    return _COUNTRIES.get(country_code.upper(), {})


class email_on_exception(object):
    def __init__(self,
                 to=("internal@odoo.com",), cc=None,
                 callback=None, reraise=True, allowed_exceptions=()):
        self.to = to
        self.cc = cc
        self.callback = callback
        self.reraise = reraise
        self.allowed_exceptions = allowed_exceptions

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(this, *args, **kwargs):
            try:
                return fn(this, *args, **kwargs)
            except self.allowed_exceptions:
                raise
            except Exception as ex:
                _logger.exception("Exception when running %r(*%r, **%r)", fn, args, kwargs)
                tb = cgitb.text(sys.exc_info())
                ex_s = exception_to_unicode(ex)
                if callable(self.callback):
                    subject, body = self.callback(this, ex_s, tb)
                else:
                    subject, body = ex_s, tb
                try:
                    env = getattr(this, 'env', None)
                    if not env:
                        # may be on a controller, use the request one
                        if request:
                            env = request.env
                    email_sys_admin(env, subject, body, email_to=self.to, email_cc=self.cc, should_raise=True)
                except Exception:
                    # ignore the errors that may occurs in mail sending
                    _logger.exception(
                        "Exception while sending traceback by email. Original Traceback:\n%s", tb)
                    pass
                if self.reraise:
                    raise
        return wrapper


def email_sys_admin(env, subject, body,
                    email_from="blackhole@odoo.com",
                    email_to=("internal@odoo.com",),
                    email_cc=None,
                    should_raise=False):
    try:
        # P3 new mail API will bail out if subject contains CRLF (13.3+)
        subject = ' | '.join(subject.splitlines()) if subject else ''

        email = env['ir.mail_server'].build_email(
            email_from=email_from,
            email_to=email_to, email_cc=email_cc,
            subject=subject, body=body,
        )
        env['ir.mail_server'].send_email(email)
    except Exception:
        if should_raise:
            raise
        _logger.exception(
            "Exception while sending email to sysadmin. Original message:%s\n\n%s", subject, body)


def guess_admin_id(cr):
    """guess the admin user id of `cr` database"""
    if not version_gte('12.0'):
        return SUPERUSER_ID
    cr.execute("""
        SELECT min(r.uid)
          FROM res_groups_users_rel r
          JOIN res_users u ON r.uid = u.id
         WHERE u.active
           AND r.gid = (SELECT res_id
                          FROM ir_model_data
                         WHERE module = 'base'
                           AND name = 'group_system')
    """, [SUPERUSER_ID])
    return cr.fetchone()[0] or SUPERUSER_ID


def get_hostname_from_ip(ip):
    normalized_ip = normalize_ip_address(ip)
    res = get_server_names_by_ip().get(normalized_ip)
    return res


def get_roles_from_ip(ip):
    res = set()

    config = get_saas_config()
    if not config:
        _logger.error("get_roles_from_ip: cannot open config.json")
    for v in config.get('pong', {}).get('workers', {}).values():
        w_ips = [v.get('public_ip'), v.get('public_ipv6', False)]
        if any([compare_ip_addresses(w_ip, ip) for w_ip in w_ips]):
            res |= set(v.get('mode', '').split())
    return res


def compare_ip_addresses(a, b):
    """ Function to compare two IPs (v4/v6) """

    if not a or not b:
        return False

    normalized_a = normalize_ip_address(a)
    normalized_b = normalize_ip_address(b)
    return hexlify(normalized_a) == hexlify(normalized_b)


def normalize_ip_address(ip):
    # we won't handle cases where the IP is invalid (e.g, 256.0.0.0 and +)
    ipv4_re = re.compile(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
    # the rational in here will be, if the IP match IPv4, then we consider it as IPv4, otherwise it is IPv6
    if ipv4_re.match(ip):
        return socket.inet_pton(socket.AF_INET, ip)
    # IPv6
    return socket.inet_pton(socket.AF_INET6, ip)


def extract_signature_from_origin(request):
    # RFC 6648 has deprecated `X-` prefix for custom parameters
    msg_signature = request.httprequest.headers.get('oe-signature', '')  # base64 encoded hmac sha256 digest
    msg_server_src = request.httprequest.headers.get('oe-server-name', '')
    msg_timestamp = request.httprequest.headers.get('oe-timestamp', '')

    # craft the body
    if request_type() == 'json':
        body = json.dumps(request_jsonrequest(), sort_keys=True)
    else:
        if request.httprequest.files:
            _logger.error('Cannot sign request with files for route "%s"', request.httprequest.path)
            raise Forbidden()
        if hasattr(request.httprequest, 'is_json') and request.httprequest.is_json:
            # http request with a mimetype that indicates JSON data (>= 16.0)
            body = json.dumps(request.get_json_data(), sort_keys=True)
        else:
            body = json.dumps(request.httprequest.form, sort_keys=True)

    message = '%s|%s|%s|%s' % (
        msg_timestamp,  # timestamp
        request.httprequest.path,  # url path
        # url query params sorted by key (I can also use args which is a multidict)
        json.dumps(dict(urllib.parse.parse_qsl(request.httprequest.query_string.decode())), sort_keys=True),
        body,  # http request body
    )
    return message, msg_signature, msg_timestamp, msg_server_src



def from_role(*roles, **options):
    """ Function to check if a controller is called by an authorized client
        - args: server roles authorized to call the controller ('*' for any role)
        - kwargs:
            - `signed` - check via a HMAC-SHA256 if the message has not been altered, true by default.
            - `mode` - origin: check the query come from the known IP and signature are embedded in the header
                               default behavior
                       params: The query param has been signed by a known server, but the query transit
                                  to another machine. Signature are checks from method args.
                                  It implies `signed` = True
            - `sig_param` - if mode is params, custom parameter name that will contain
                            the signature (default: '_ms' aka "method signature")
    """
    roles = set(roles)
    mode = options.get('mode', 'origin')
    sig_param = options.get('sig_param', '_ms')
    sig_ttl = options.get('sig_ttl', TIMEOUT)

    def decorator(func):

        @wraps(func)
        def wrapper(*args, **kw):
            saas_config = get_saas_config()
            workers = saas_config.get('pong', {}).get('workers', {})

            if mode == 'origin':
                fltr = lambda modes: [(w['public_ip'], w.get('public_ipv6', False))
                                      for w in workers.values()
                                      if '*' in modes or len(set(w['mode'].split()) & modes)]
                # 1. check if the source ip addr has one of the authorized roles
                auth_ips = fltr(roles)
                # Flatten the list of tuples of IP addresses
                auth_ips = {ip for ips_tuple in auth_ips for ip in ips_tuple}
                if not any(compare_ip_addresses(request.httprequest.remote_addr, ip) for ip in auth_ips):
                    raise Forbidden()

                # 2. Nginx replaces trusted real ips in X-Forwarded-For by client one.
                # As the load-balancer are trusted, this lead to having the client ip twice in the
                # access route. Deduplicate them.
                access_route = set(request.httprequest.access_route)
                if len(access_route) != 1:
                    # should only pass through 1 proxy (nginx)
                    raise Forbidden()

                message, msg_signature, msg_timestamp, msg_server_src = extract_signature_from_origin(request)
            else:
                message, msg_signature, msg_timestamp, msg_server_src = extract_signature_from_kwargs(
                    request.httprequest.path, kw, sig_param, sig_ttl
                )
            # 3. Check message integrity
            if options.get('signed', True) or msg_signature or mode != 'origin':
                try:
                    # if any of the required parameters to check the message integrity is missing, we raise forbidden
                    if not msg_server_src or not msg_signature or not msg_timestamp:
                        raise Forbidden()

                    server_config = workers.get(msg_server_src, {})

                    # base64 encoded 256 bit key
                    server_http_secret = server_config.get('http_shared_secret')
                    if not server_http_secret:
                        _logger.error('Unable to retrieve http secret for server "%s"', msg_server_src)
                        raise Forbidden()

                    if mode == 'origin':
                        # check if the request's source ip addr match the server public ips (or localhost)
                        authorized_ips = [
                            server_config.get('public_ip'),
                            server_config.get('public_ipv6'),
                            '127.0.0.1',
                            '::1',
                        ]

                        if not any(compare_ip_addresses(request.httprequest.remote_addr, ip) for ip in authorized_ips):
                            _logger.error(
                                'HTTP request validation failed, source ip does not match server ips %r for route "%s"',
                                authorized_ips,
                                request.httprequest.path)
                            raise Forbidden()
                    else:
                        # Check server has the proper roles
                        if not ('*' in roles or len(set(server_config['mode'].split()) & roles)):
                            raise Forbidden()

                    if not check_message_signature(message, msg_signature, msg_timestamp, server_http_secret, sig_ttl):
                        _logger.error('HTTP request validation failed due to invalid signature for route "%s"',
                                      request.httprequest.path)
                        raise Forbidden()

                except Exception:
                    _logger.exception('Error while validating integrity of http request for route "%s"',
                                      request.httprequest.path)
                    raise Forbidden()
            elif options.get('signed') is not False:
                _logger.info('Message integrity of route "%s" is not checked.', request.httprequest.path)
            return func(*args, **kw)
        return wrapper

    return decorator


def has_role(role):
    saas_config = get_saas_config()
    return role in saas_config.get('me', {}).get('mode', '').split()

def get_saas_config():
    try:
        with open(os.path.join(os.path.expanduser('~/config'), 'config.json')) as c:
            saas_config = json.load(c)
    except (IOError, ValueError):
        saas_config = {}
    return saas_config

def get_my_server_config():
    saas_config = get_saas_config()
    return saas_config.get('me', {})

def get_server_config(server_name):
    saas_config = get_saas_config()
    return saas_config.get('pong', {}).get('workers', {}).get(server_name, {})

def get_server_name():
    return get_my_server_config().get('name', False)

def get_server_names_by_ip():
    saas_config = get_saas_config()
    ret = {}
    for h in saas_config.get("pong", {}).get("workers", {}).values():
        public_ipv6 = h.get("public_ipv6", False)
        if public_ipv6:
            ret[normalize_ip_address(public_ipv6)] = h["name"]
        ret[normalize_ip_address(h["public_ip"])] = h["name"]
    return ret

def str2bool(s, default=None):
    s = ustr(s).lower()
    y = 'y yes 1 true t on'.split()
    n = 'n no 0 false f off'.split()
    if s not in (y + n):
        if default is None:
            raise ValueError('Use 0/1/yes/no/true/false/on/off')
        return bool(default)
    return s in y

def baseN(num, b, numerals=string.digits + string.ascii_letters):
    assert num >= 0
    assert 1 < b <= len(numerals)
    return ((num == 0) and "0") or (baseN(num // b, b, numerals).lstrip("0") + numerals[num % b])

def hash62(data):
    h = hashlib.md5(str.encode(data))
    return baseN(int(h.hexdigest(), 16), 62)

def wrap_module(module, attr_list):
    """Helper for wrapping a package/module to expose selected attributes

       :param Module module: the actual package/module to wrap, as returned by ``import <module>``
       :param iterable attr_list: a global list of attributes to expose, usually the top-level
            attributes and their own main attributes. No support for hiding attributes in case
            of name collision at different levels.
    """
    attr_list = set(attr_list)
    class WrappedModule(object):
        def __getattr__(self, attrib):
            if attrib in attr_list:
                target = getattr(module, attrib)
                if isinstance(target, types.ModuleType):
                    return wrap_module(target, attr_list)
                return target
            raise AttributeError(attrib)
    # module and attr_list are in the closure
    return WrappedModule()


def recaptcha_v3_get_score(secret_key, token):
    """Retrieve and log the score for a Google reCAPTCHA token for
       the current requesting IP (`httprequest.remote_addr`)

        :param string secret_key: The secret key for the Google reCATPCHA API
        :param string token: The Google reCATPCHA token for which you want to retrieve the score.

        :return float|string|boolean:
            When everything is working, the score is returned as a float,
            when the call to the Google reCATPCHA times out, the 'timeout' string is returned,
            when we were unable to retrieve the token (non existing token, already consumed one) `False` is returned
    """
    google_recaptcha_url = 'https://www.recaptcha.net/recaptcha/api/siteverify'

    ip_addr = request.httprequest.remote_addr
    try:
        r = requests.post(google_recaptcha_url, {
            'secret': secret_key,
            'response': token,
            'remoteip': ip_addr,
        }, timeout=2) # it takes ~50ms to retrieve the response
    except requests.exceptions.Timeout:
        _logger.error("reCAPTCHAv3 verification timeout for ip address %s", ip_addr)
        return 'timeout'

    result = r.json()
    # if Google reCAPTCHA replies with a `success`, we return the score and print a line in the log
    if result['success'] is True:
        score = result.get('score', False)
        _logger.info("reCAPTCHAv3 verification for ip address %s succeeded with score %f.", ip_addr, result['score'])
        return score

    # if Google reCATPCHA doesn't reply with a `success`, we return False and print the error code on the logs
    _logger.error(
        "reCAPTCHAv3 verification for ip address %s failed error codes %r. token was: [%s]",
        ip_addr,
        result.get('error-codes', 'None'),
        token
    )
    return False

def replace_session(user_id, impostor_info=None):
    if version_gte("saas~15.3"):
        request.update_env(user=user_id)
    else:
        request.uid = user_id
        vars(request).pop('env', None)  # env has to be re-created with the new user
    request.session.uid = user_id
    user = request.env['res.users'].browse(user_id)

    # Manual (cross-version) session rotation: destroy old session + new sid
    root.session_store.delete(request.session)
    request.session.sid = root.session_store.generate_key()

    # Invalidate + regenerate token to match new uid and sid
    if version_gte("saas~16.4"):
        user.env.registry.clear_cache()
    elif version_gte('14.0'):
        user.clear_caches()
    else:
        user._invalidate_session_cache()

    request.session.session_token = user._compute_session_token(request.session.sid)

    if impostor_info:
        request.session.impostor_info = impostor_info

    # update the login date if not an impostor (/_s2s/auth/check-token always returns impostor info)
    if not impostor_info or user.oauth_uid == str(impostor_info.get('id')):
        user._update_last_login()

def check_oauth2_token(token):
    """ Send a request to oauth2 server to validate the oauth2 access token

        :param str token: oauth2 access token to validate

        :return dict:
            the json response of the oauth server
            that contains the oauth_uid and the impostor info

        :raises: werkzeug.exceptions.Forbidden if the token is not valid or has expired,
                 or if the database could not be found
    """
    db_info = metabase.get(request.session.db)
    if not db_info:
        raise Forbidden('Invalid database')
    data = {'params': {'token': token, 'db_uuid': db_info.uuid}}
    ip = request.httprequest.remote_addr
    response = requests.post(OAUTH_BASE_URL + '/_s2s/auth/check-token', auth=SaasAuth(), json=data, timeout=TIMEOUT)
    response = response.json()
    result = response.get('result', {})
    if response.get('error') or not isinstance(result, dict) or not result.get('oauth_uid'):
        _logger.warning('check-token failed with token=%s ip=%s error=%s', token[:8], ip, response.get('error', result))
        raise Forbidden('Invalid or missing token')

    _logger.info('check-token passed with token=%s oauth_uid=%s ip=%s', token[:8], result.get('oauth_uid'), ip)
    return result


def check_oauth2_tokens(token_per_db_uuid):
    """ Send a request to oauth2 server to validate multiple oauth2 access tokens

        :param dict token_per_db_uuid: a mapping of [str] token for each [str] db UUID

        :return dict:
            the json response of the oauth server that contains the oauth_uid and impostor info
            for each database uuid
    """
    ip = request and request.httprequest.remote_addr or 'unknown'
    data = {'params': {'token_per_db_uuid': token_per_db_uuid}}
    response = requests.post(OAUTH_BASE_URL + '/_s2s/auth/check-tokens', auth=SaasAuth(), json=data, timeout=TIMEOUT)
    response = response.json()

    error = response.get('error', False)
    if error:
        _logger.warning('check-token failed with tokens=%r ip=%s error=%s',
                      {u: t[:8] for u, t in token_per_db_uuid.items()}, ip, error)
        raise Exception('Error while checking tokens')

    result = response.get('result', {})
    for db_uuid, token in token_per_db_uuid.items():
        res = result.get(db_uuid, False)
        if res:
            _logger.info('check-token passed with db=%s token=%s oauth_uid=%s ip=%s', db_uuid, token[:8],
                         res.get('oauth_uid'), ip)
        else:
            _logger.warning('check-token failed with db=%s token=%s ip=%s error=invalid token', db_uuid, token[:8], ip)

    return result


def _get_user_from_oauth_uid(oauth_uid):
    env = request.env
    ResUsers = env['res.users']

    if not has_role('trial'):
        _logger.info("_get_user_from_oauth_uid: _get_user_from_oauth_uid should only be used on trial servers route=%s", request.httprequest.path)
        raise werkzeug.exceptions.Forbidden()

    domain = [('oauth_uid', '=', oauth_uid)]
    if 'oauth_provider_id' in ResUsers._fields:
        oe_provider = env.ref('auth_oauth.provider_openerp', raise_if_not_found=False)
        provider_dom = [('oauth_provider_id', '=', False)]
        if oe_provider:
            provider_dom = expression.OR([provider_dom, [('oauth_provider_id', '=', oe_provider.id)]])
        domain = expression.AND([domain, provider_dom])

    oauth_user = ResUsers.sudo().search(domain, limit=1)
    return oauth_user


@contextmanager
def replaced_session(uid, impostor_info=None):
    # backup current session information
    old_req_uid = request.uid
    old_req_session_uid = request.session.uid
    old_req_impostor_info = request.session.get('impostor_info')
    had_impostor_info = bool(old_req_impostor_info)
    if had_impostor_info:
        old_req_impostor_info = dict(old_req_impostor_info)

    try:
        # replace session information
        if version_gte("saas~15.3"):
            request.update_env(user=uid)
        else:
            request.uid = uid
        request.session.uid = uid
        if impostor_info:
            request.session.impostor_info = impostor_info

        # call with block
        yield
    finally:
        # restore previous session information
        if version_gte("saas~15.3"):
            if old_req_uid is None:
                # NOTE: We have to create a new env from scratch as `request.update_env(user=None)`
                # will not update the user (`None` being the default value it will fallback
                # to the existing env user, i.e. will not change it)
                request.env = api.Environment(
                    request.env.cr, old_req_uid, request.env.context, request.env.su
                )
                threading.current_thread().uid = old_req_uid
            else:
                request.update_env(user=old_req_uid)
        else:
            request.uid = old_req_uid
        request.session.uid = old_req_session_uid
        if had_impostor_info:
            request.session.impostor_info =  old_req_impostor_info
        else:
            request.session.pop('impostor_info', None)


@contextmanager
def replaced_database(db_name, uid=SUPERUSER_ID, oauth_uid=None, oauth_token=None):
    """
    Force the request.session.db and request.session.uid,
    so that any controller method will use them within this context.
    When the context is exited, the previous db and uid are restored.
    At the end of the block, changes will be committed unless an exception is raised.
    However, any changes that would be uncommitted before the block will be rollbacked.

    When oauth_token is provided, uid is ignored, and the user identifier will be retrieved from the token.
    """
    def _replace_env(_db_name, _uid=None, _oauth_uid=None, _oauth_token=None, _env=None):
        # force the database name
        request.session.db = _db_name
        if hasattr(request, '_db'):
            request._db = _db_name
        threading.current_thread().dbname = _db_name or None

        if version_gte("saas~15.3"):
            if request.env:
                if not request.env.cr.closed:
                    request.env.cr.close()
                request.env = None

            # with http-ocalpyse, registry/env are not automatically instanciated
            # on demand anymore, and cr/uid/context depends on env, so we have to
            # manage it ourselves
            # Note: Now that auth="none" request will have env=None, accessing
            # request.cr will crash with NoneType has no attribute "cr" if we
            # don't create a valid env and registry beforehand (same applies for
            # request.uid and request.context)
            if not _db_name:  # reset from auth=none
                request.db = None
                request.registry = None
                request.env = None
            else:
                request.db = _db_name
                request.registry = RegistryManager.get(_db_name)
                env_cr = request.registry.cursor()
                env_uid = _env.uid if _env else None
                env_context = _env.context if _env else request.session.context
                request.env = api.Environment(env_cr, env_uid, env_context)

        else:
            # < saas~15.3: close cursor and invalidate env, they will be auto.
            #  re-created when needed.
            if request._cr:
                request._cr.close()
                request._cr = None

            # clear the environment
            if hasattr(request, '_env'):
                # >= saas~13: env is cached in _env
                request._env = None
            else:
                # < saas~13: env is a lazy_property
                vars(request).pop('env', None)

        # set the new uid
        if _oauth_token:
            _uid = retrieve_credentials_from_token(oauth_token=_oauth_token)[0]
        elif _oauth_uid:
            # Note: _get_user_from_oauth_uid() needs a valid request.env
            _user_id = _get_user_from_oauth_uid(_oauth_uid)
            if not _user_id:
                _logger.info("replaced_database: No such user: db=%s oauth_uid=%s", _db_name, _oauth_uid)
                raise werkzeug.exceptions.Forbidden("No such user")
            _uid = _user_id.id
        if _uid:
            if version_gte("saas~15.3"):
                request.update_env(user=_uid)
            else:
                request.uid = _uid
            request.session.uid = _uid
        elif not _db_name:
            # always reset session uid when switching to nodb
            request.session.uid = None

    bkp_db_name = request.session.db
    bkp_uid = request.session.uid
    bkp_env = request.env if version_gte("saas~15.3") else None
    try:
        _replace_env(db_name, _uid=uid, _oauth_uid=oauth_uid, _oauth_token=oauth_token)
        yield
        request.env.cr.commit()
    finally:
        _replace_env(bkp_db_name, _uid=bkp_uid, _env=bkp_env)


def oauth2_token_to_user(token):
    token_result = check_oauth2_token(token)
    oauth_uid = token_result.get('oauth_uid')
    impostor_info = token_result.get('impostor_info')

    oauth_user = _get_user_from_oauth_uid(oauth_uid)
    if not oauth_user:
        ip = request.httprequest.remote_addr or 'unknown'
        _logger.warning('oauth2_token_to_user: user not found oauth_uid=%s ip=%s', oauth_uid, ip)

    return oauth_user, impostor_info


def oauth2_token_auth(mandatory=True):
    """
    Decorates a route that supports authentication via an oauth2 token.
    The correct user will be impersonated for this call, and restored after.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, oauth_token=None, **kw):
            ip = request.httprequest.remote_addr

            if not has_role('trial'):
                _logger.info("oauth2_token_auth: oauth2_token_auth should only be used on trial servers route=%s ip=%s", request.httprequest.path, ip)
                raise werkzeug.exceptions.Forbidden()

            if not oauth_token:
                if mandatory:
                    _logger.info("oauth2_token_auth: mandatory token not provided route=%s ip=%s", request.httprequest.path, ip)
                    raise BadRequest()
                else:
                    return func(*args, **kw)

            oauth_user, impostor_info = oauth2_token_to_user(oauth_token)

            if oauth_user:
                with replaced_session(oauth_user.id, impostor_info):
                    return func(*args, **kw)
            else:
                return func(*args, **kw)

        # update function signature to insert `oauth_token param at the end
        # of parameters or just before "**kwargs" (if present), so that
        # dispatcher does not ignore it
        func_sign = inspect.signature(func)
        func_params = list(func_sign.parameters.values())
        oauth_token_param = inspect.Parameter(
            name="oauth_token", kind=inspect.Parameter.POSITIONAL_OR_KEYWORD, default=None,
        )
        has_kwargs = bool(func_params and func_params[-1].kind == inspect.Parameter.VAR_KEYWORD)
        new_func_params = list(func_params)
        new_func_params.insert(len(new_func_params) - int(has_kwargs), oauth_token_param)
        wrapper.__signature__ = func_sign.replace(parameters=new_func_params)

        return wrapper

    if hasattr(mandatory, '__call__'):
        # Decorator used without argument, receives func directly
        func, mandatory = mandatory, True
        return decorator(func)
    return decorator


@oauth2_token_auth
def retrieve_credentials_from_token():
    """
    Retrieves the uid and impostor_info from inside a decorated method
    """
    return request.session.uid, request.session.impostor_info


def staff_action_authorized(action, reason=None, impostor_info=None, impersonated_user_id=None):
    """Check staff access for the specified action and impersonated user.
    :param action: action that the user is trying to perform (see openerp.enterprise.spoof model)
    :param reason: information about action being performed and possible justification
    :param impostor_info: information about user doing the impersonation (retrieved from session if not provided)
    :param impersonated_user_id: Id of the user being impersonated (retrieved from request/session if not provided)
    """
    db_name = request.session.get('db')
    db = metabase.get(db_name)
    impostor_info = impostor_info or request.session.impostor_info or {}
    if impersonated_user_id:
        impersonated_user = request.env['res.users'].browse(impersonated_user_id)
    else:
        impersonated_user = request.env.user or request.env['res.users'].browse(int(request.session.uid))
    # Read the user information as sudo. We do not want to be blocked by record rules
    # checking if we are reading a user in our own company since we might not have a
    # user in our env, this leads to crashes.
    impersonated_user = impersonated_user.sudo()

    spoof_info = {
        'ip': request.httprequest.environ['REMOTE_ADDR'],
        'login': impostor_info.get('login'),
        'user_id': impostor_info.get('id'),
        'action': action,
        'reason': reason or "",
        'db_uuid': db.uuid,
        'remote_uid': impersonated_user.id,  # remote from the PoV of accounts.odoo.com
        'remote_login': impersonated_user.login,
    }
    if not impostor_info:
        # customer is doing the action
        spoof_info['login'] = request.env.user.login
    params = {'params': spoof_info}
    granted = False
    try:
        response = requests.post(
            '%s/odoo-enterprise/staff-access' % OAUTH_BASE_URL, auth=SaasAuth(), json=params, timeout=TIMEOUT
        ).json()
        granted = response.get('result')
        return granted
    finally:
        _logger.info("staff-access: %s %s by %s (as %s), reason: %s, db_uuid: %s",
            granted and "GRANTED" or "REFUSED",
            action,
            impostor_info.get('login') or request.env.user.login,
            impersonated_user.login,
            reason or "",
            db.uuid,
        )


def get_ip_info(api_token, ip_addr=None):
    try:
        ip_address = ip_addr or request.httprequest.remote_addr
        auth_headers = {"Authorization": "Bearer %s" % api_token}
        req = requests.get('https://ipinfo.io/%s/json' % ip_address, headers=auth_headers, timeout=2)
        # if we consumed the number of allowed requests
        if req.status_code == requests.codes.TOO_MANY_REQUESTS:
            return {'error': 'TOO_MANY_REQUESTS'}

        # if service down
        if req.status_code == requests.codes.SERVICE_UNAVAILABLE:
            return {'error': 'SERVICE_UNAVAILABLE'}

        ip_info = req.json()

        if 'ip' not in ip_info:
            _logger.error("ipinfo lookup failed for ip %s: %r", ip_address, ip_info)
            return {'error': 'LOOKUP_FAILED'}

        _logger.info("ipinfo lookup for ip %s: %r", ip_address, ip_info)
        return ip_info
    except requests.exceptions.Timeout:
        _logger.error("ipinfo timeout for ip address %s", ip_address)
        return {'error': 'TIMEOUT'}

    return {'error': 'UNKNOWN'}

filestore = partial(os.path.join, os.path.expanduser('~/filestore'))
